"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                       PYTHON FUNCTIONS                                      ||
      +=================-=========-=========-=========*=========-=========-=========-=================+

   The Python interpreter has a number of functions and types built into it that are always available--i.e.,
you do NOT need to import them. Here are some of the most commonly used built-in functions.

	                                 == Some Useful Built-in Functions ==

    abs()           absolute value for both int and float types

    bin()           print (bin(17))     # output: 0b10001
                    print (bin(17)[2:]) # output: 10001

    bool()          print(bool('A'), bool(-1), bool(0), bool([])) # output: True True False False

    chr()           print(chr(48), chr (49), chr(65), chr(66), chr(97), chr(122))
                    # output: 0 1 A B a z

    dir             print(dir(str)) prints all function names associated with a data type. Use with help().

    enumerate()     print(list(enumerate(['a', 'b', 'c']))) # output: [(0, 'a'), (1, 'b'), (2, 'c')]

    eval()          print(eval ('2 + 2')) # output: 4

    float()         casts to a floating point number.

    help()          print(help (round)), explains a function

    hex()           print (hex(17))     # output: 0x11
                    print (hex(17)[2:]) # output: 11

 	input()         x = int(input('enter an integer'))

    int()           casts to an int type

    isinstance()    print(isinstance(7, (int, float))) # output: True

    len()           print (len(['A', 'B', 'C',])) # output: 3

    list()          casts to list type

    max()           print (max([1,2,3,4,])) # output: 4

    min()           print (min([1,2,3,4,])) # output: 1

    oct()           print (oct(17))     # output: 0o21
                    print (oct(17)[2:]) # output: 21

    ord()           print(ord('0'), ord ('1'), ord('A'), ord('B'), ord('a'), ord('z'))
                    # output: 48 49 65 66 97 122

    pow()           print(pow(4, 3), 4**3) # output: 64 64

    print()         print ('Hello', 5) # output = Hello 5
                    print ('int = %5d, float = %7.2f, string = %7s'%(12, 34.567, 'money'))
                    # output: int =    12, float =   34.57, string =   money [Notice the spacing.]

    range           print(list(range(4)))   # output: [0, 1, 2, 3]

    round()         print(round(1.46), round(1.46, 1)) # output: 1 1.5

    set()           cast to a set type, or create an empty set: x = set()

    sort()          Lst = [4,1,3,2,]
                    Lst.sort()
                    print(Lst)   # output: [1, 2, 3, 4] See sorted().

    sorted()        print(sorted([4,1,3,2,])) # output: [1, 2, 3, 4]

    str()           casts to a string type

    sum()           print(sum([4,1,3,2,]))  # output: 10

    tuple()         cast to a tuple type.

    type()          print(type('ABC'))           # output: <class 'str'>

    zip()           print(list(zip([1,2,3], ['A', 'B', 'C',])))  # output: [(1, 'A'), (2, 'B'), (3, 'C')]
#-------------------------------------------------------------------------------------------------------------

   The range function, mainly used with for loops, looks like this:

                             range (start, stop, step).

      If start is omitted, it defaults to 0. If step is omitted it defaults to 1.

   When would you need to step? Most likely it would be when you are working with floats. Surprise! The range
function does NOT work for floats. You must write your own function. Here is mine, which uses a yield instead
of a return. What is a yield? It returns the next value of the loop, every time you call the function. Other-
wise, the loop would start over every time you called the function. Place this function in your personal
library in case you ever need it.

                    +-------------------------------------+
                    |  def frange(start, stop, step = 1): |
                    |      i = start                      |
                    |      while i < stop:                |
                    |         yield i # <-- not return    |
                    |         i += step                   |
                    +-------------------------------------+

                      def main():
                          from math import pi
                          for t in frange (0, 2*pi, 0.1):
                              print(t)
#-------------------------------------------------------------------------------------------------------------
   What built-in functions are in the math module? We can find out using the dir() function:

                import math
                print (dir(math)) #<--Lists all math functions

       Output: ['__doc__', '__loader__', '__name__', '__package__', 'acos', 'acosh', 'asin',
                'asinh', 'atan', 'atan2', 'atanh', 'ceil', 'copysign', 'cos', 'cosh', 'degrees',
                'e', 'erf', 'erfc', 'exp', 'expm1', 'fabs', 'factorial', 'floor', 'fmod',
                'frexp', 'fsum', 'gamma', 'hypot', 'isfinite', 'isinf', 'isnan', 'ldexp',
                'lgamma', 'log', 'log10', 'log1p', 'log2', 'modf', 'pi', 'pow', 'radians',
                'sin', 'sinh', 'sqrt', 'tan', 'tanh', 'trunc']

    But, we can do better by executing this line:

                             help('math') or in particular help('atan2')

    How many modules are there? We can find out (almost) by executing these lines:

                    import sys
                    print(sorted(list(sys.modules.keys())))
#-------------------------------------------------------------------------------------------------------------

Every user-built function begins with "def", meaning "definition." Here is an example:

                                    def printMessage(msg):
                                        print(msg)

   Immutables are passed by value in Python, but mutables are passed by reference--e.g., strings and numbers
are passed to functions by value, while lists and tuples are passed by reference. This is usually what you
want, because working with copies of huge lists or dictionaries is too time-and-memory consuming. This means
you don't need to return a changed list that was passed to a function. Some programmers feel that you should
return it anyway, to make the program easier to follow.

   However, there are three ways to force a function to work with a copy of a list.

      1. doit(a, tuple(L)) # make L immutable, but this is too extreme.
      2. doit(a, L[:])     # pass a copy (nice), but fails with lists of lists.
      3. def doit (a, L)
           from copy import deepcopy
           M = deepcopy(L) # always works even for lists of lists (best).

   Return is optional, but without an explicit return, the value None is returned.

   Functions can have default values. Note that all of the variables to the right of the first default value
must also have default values.
                                    def doit(a, b = 2, c = 3):
                                    ...

   Besides the usual "passing by position", we can use keyword arguments, in which case the positions of the
parameters are irrelevant.
                                    doit(a = 1, c = 3, b = 2)

   Functions may contain other functions. Why would we want this? Because when we copy a function, we pick up
the utility functions it calls. This keeps the tools in the same box as the function that uses them.

   Warning: Beware of accidentally switching the order of passed parameters--e.g., passing (row, col) and
receiving (col, row). This is a difficult-to-find error.

    It is possible for a function to return more than one value, by returning the values in a list (surrounded
by square brackets) or a tuple (parentheses are optional).

   Using the star operator(*) we can allow functions to accept any number of arguments (polymorphism) and pack
them into a tuple. In Python, you can NOT have two functions with the same names, even if they receive
different numbers of parameters (lack of polymorphism).

                    def doit(*args):
                        for n in args:
                            print(n, end = ',')
                    def main():
                        doit(1,2,3) # output: 1,2,3,

    Recall functions are objects. Consequently, they can have class (static) variables, instance variables,
and they can have their own methods (functions inside the functions). Below we see an example of a class
(static) variable. When do we need static variables in a function? Answer: When we want a function to
remember its previous state when called later.

            def doit(x):
              x += 1
              doit.y += 1
              print(x, doit.y)
            #---------------------------
            doit.y = 0 # static variable initialization must appear BELOW the function it references.
            #---------------------------
            def main():
                doit(0) # Output: 1 1
                doit(0) # Output: 1 2

  Avoid defaulting to a mutable expression, because this expression is given only one place in memory, and is
not allowed to change its location. Consequently, when you call the function again, the default value is NOT
used to re-initialize.

                            def doit1(x = None): # Correct
                                if x == None:
                                    x = []
                                x.append(1)      # Recall: x.append(1) appends in place and returns None.
                                return x


                            def doit2(x = []):   # A BIG mistake
                                x.append(1)
                                return x

                            print ( doit1() ) # Output: [1]
                            print ( doit1() ) # Output: [1]
                            print ( doit2() ) # Output: [1]
                            print ( doit2() ) # Output: [1, 1] <-- WHAT?!
#-------------------------------------------------------------------------------------------------------------
