"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                     GENERATORS and YIELD                                    ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+

   The instruction "yield" means to return, but not terminate. The result is that a function returns a
sequence of values. The values are collected by the receiver. A useful example is frange (float range).

                    +-------------------------------------+
                    |  def frange(start, stop, step = 1): |
                    |      i = start                      |
                    |      while i < stop:                |
                    |         yield i # <-- not return    |
                    |         i += step                   |
                    +-------------------------------------+

                      def main():
                          from math import pi
                          for t in frange (0, 2*pi, 0.1):
                              print(t)

   Of course, we could just collect them all in the function and return a list or tuple of values. but that
may take too much memory, and at least requires some memory management.

   You can make a list comprehension into a generator by using parentheses, instead of square brackets.

        dataList = [x for x in 'ABC'] # This is a list comprehension. Note the brackets.
        print(dataList)               # ['A', 'B', 'C', 'D']

        dataList = (x for x in 'ABC') # This is a generator. Note the parentheses.
        print(dataList)               # <generator object <genexpr> at 0x023BB788>

        print(next(dataList))         # A [This will also work: print(dList().__next__()).]
        print(next(dataList))         # B
        print(next(dataList))         # C
#       print(next(dataList))         # Error: StopIteration [, because the generator is used up.]

        print(dList())                # <generator object dList at 0x0220EA30>
        print(next(dList()))          # A
        print(next(dList()))          # A <--Surprise, the dList() function is reset.
        print(list(dList())[1])       # B <--This defeats the purpose of using a generator, instead of a list.

        for x in dList():
            print (x, end = ',')      # A,B,C,
        print()
        for x in dList():
            print (x, end = ',')      # A,B,C, The generator is NOT used up.