"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                               PYTHON DEMO Tk-GRAPHICS PROGRAM                               ||
      ||                                (canvas drawing and animation)                               ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   This is a working Python/Tk graphics program. Tk is a toolkit for GUI programming.  A widget is a GUI
object (window, button, scrollbar, menu, etc.). Tk has a general-purpose drawing widget (window) called
Canvas.  Widgets are placed on the screen and sized by a geometry manager. There are three different kinds of
managers: pack, grid, and place. We will use only pack in this section.

   Tk guarantees to support the three font families named 'Courier' (a mono-spaced font), 'Times' (a serifed
font), and 'Helvetica' (a sans-serif font). To write to the screen, we use the "create_text()" command that
has many options. The general case follows.

   canvas.create_text(col, row,
                      text = 'your text string goes here',
                      fill = 'color',
                      font = ('name': Times, Courier, Helvetica',
                               size: 20 is big,
                              'bold', 'italic', 'normal', 'roman',
                              'underline', 'overstrike') )

   For example, here is a line from a working program:

    canvas.create_text (11, 17, text = 'X', fill = 'red', font = ('Helvetica', 20,'bold') ).

   You can combine the font name and its size as 'Helvetica 20'. The color attribute can be a special name,
like 'RED' or an RGB number like '#000000', where the digits are hexadecimal--e.g.,

     '#ffffff' = white,   '#ff0000' = red,   '#00ff00' = green,   '#0000ff' = blue,   '#000000' = black.

   Since the colors can be mixed, there are 16^6 = 16777216 possible different colors. A few of the special
names you can use are black, grey, white, red, green, blue, orange, brown, cyan, magenta, yellow, pink, tan,
etc. Many of these colors allow the digits 1,2,3,4 after their names, like blue, blue1, blue2, blue3, and
blue4.

   Note well: the graphics images are buffered. This means they do NOT appear on the screen when you create
them. To get them to appear, you must use canvas.update(). This instruction is sometimes followed by the sleep
instruction.

                canvas.update()      # Make all previous changes to the canvas.
                sleep (1)

  If you draw too many images, the computer will begin to slow down. Thus, the clear command (discussed below)
is useful. Good luck.
"""

##############################################<START OF PROGRAM>##############################################

def setUpCanvas(): # These are the required magic lines to enter graphics mode.
    from tkinter import Tk, Canvas, YES, BOTH
    root=Tk()
    root.title("Virus: A Tk/Python Graphics Program")
    canvas = Canvas(root, width = 1270, height = 780, bg = 'BLACK')
    canvas.pack(expand = YES, fill = BOTH)
    canvas.bell(0)                         # ring bell (at least in a Tk-graphics Windows program).
    return canvas, root
#-------------------------------------------------------------------------------------------Tk Graphics demo--

def runGraphicsCode():

#---QUESTION: Why place a function inside another functions, when it will work well outside of the enveloping
#             function?
#   ANSWER: This keeps the functions together. But note: the function must be placed ABOVE its call.

    def disk(x, y, r, kolor): #create a solid disk with center (x,y), radius r, and color named kolor.
      return canvas.create_oval(x-r,y-r,x+r,y+r, width = 1, fill = kolor)
#   ---------------------------------------------------------------------
#--1. Draw a line of thickness 2 (1 = default) and color green. Note that the name below, obj1, is optional.
#    You will only use it if you want to reference the object (line) later.

    obj1 = canvas.create_line(100,100,200,200, width = 2, fill = 'GREEN')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--2. Draw a circle of thickness 3 (default = 1) filled with blue color. The coordinates are opposite-box-
#     coordinates.

    obj2 = canvas.create_oval(100,100,200,200, width = 3, fill = 'BLUE')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--2b. Draw a solid circle of radius r at position (x,y).

    disk(200, 300, 100, 'magenta') # the 'objx =' is always optional.
    canvas.update()      # make all previous changes to the canvas
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--3. Draw a rectangle of thickness 4 (default = 1) filled with yellow color.

    obj3 = canvas.create_rectangle(500,500,700,700, width = 4, fill = 'YELLOW')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--4. Plot a point of thickness 2 (default = 1) colored blue

    obj4 = canvas.create_rectangle(700, 200, 700, 200, width = 2, fill = 'BLUE')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--5a. Write text to the video screen.

    obj5 = canvas.create_text(259, 280,text = "Hello", fill = 'GREEN', font = 'times')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--5b. Reconfigure obj5 (change the text during the run). Without its optional name (obj5), we would NOT be
#  able to reconfigure the line.

    canvas.itemconfigure(obj5, text='Hey!', font='Courier 48')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--6. Draw a polygon. (x0, y0, x1, y1, x2, y2) will give a triangle.

    obj6 = canvas.create_polygon(300, 300, 400, 400, 300, 400, fill = 'GREEN', outline = 'BLACK', \
                             width = 5, tag = 'Tag1')
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--7. Embed an Image. (First, download a GIF file and save it in a directory.) PhotoImage supports ONLY
#     GIF graphic file formats, unless you install an external library.  And no, you cannot resize a PPM
#     image. This code will not run without the file gifImage3.gif in your directory.

    photo = PhotoImage(file = 'g:\Garfield.gif')
    print("w = ",  photo.width(), ' c = ',  photo.height())
    obj7  = canvas.create_image(250, 0, image = photo, anchor = NW) # embed photo
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--8. Move images from the graphics memory buffer to the screen.

    canvas.update()      # make all previous changes to the canvas
    sleep(1.0)           # pause 1-1/2 seconds.
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--9. Clear the screen.

    canvas.delete('all') # or canvas.delete(img?) for each image # <-- Method 1 (best)
    canvas.create_rectangle(0, 0, 1270, 780, fill = 'BLACK')     # <-- Method 2 will also clear.
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#-10. Fill the screen with Xs (characters that can be targets).

    ch = 'X'
    for row in range (1, 700, 20):
        for col in range (1, 1200, 20):
            canvas.create_text(col, row, text = ch, fill = 'GREEN', font = ('Helvetica', 20, 'bold') )
#-------------------------------------------------------------------------------------------Tk Graphics demo--





#--11.Mouse Buttons and keys. Bind the mouse buttons to the canvas.
#  Notes:  The left  button on the mouse is referenced as '<Button-1>'.
#          The right button on the mouse is referenced as '<Button-3>'.
#
#          The 'q'      key is referenced as '<q>'.
#          The up-arrow key is referenced as '<Up>'.
#
#   To make the program respond to a left-button click (and run a function), we bind the left Mouse-button to
#   the canvas (which is placed on the program's window (called root)) with
#
#                     root.bind('<Button-1>', function )
#
#   --------------------------------------------------------------------------------------------
#   Here are three more functions inside another function. Below, event.x and event.y are the mouse
#   coordinates. Because I wanted "Goodbye" to appear at location (200,200), I did not use (event.x, event.y).

    def disk(x, y, r, kolor):
      return canvas.create_oval(x-r,y-r,x+r,y+r, width = 1, fill = kolor)

    def leftBtn(evnt): # Yes a function inside a function.
       canvas.create_text(evnt.x, evnt.y,text = 'Hold and move', fill = 'GREEN', font = 'times')

    def rightBtn(evnt):
       canvas.create_text(200, 200,text = 'Goodbye', fill = 'RED', font = 'times')
#   -------------------------------------------------------------------------------------------

    root.bind('<Button-1>' , leftBtn )
    root.bind('<Button-3>' , rightBtn)
    root.bind('<B1-Motion>', leftBtn )  # <-- Look! Mouse drag motion.

#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--12.ANIMATION--Type I. This is one of two different methods to achieve animation using Tk. Here we
#     repeatedly print, erase, move, and print again. The screen MUST be updated after every movement. This
#     method gives smoother animation compared to using the after() command. Also, since objects are placed on
#     top of each other, the movement of an object over another object does not erase the underlying object.
#     How cool is that?

    canvas.create_rectangle(100,100,700,700, width = 4, fill = 'pink')
    ball1 = disk (200 , 200,  10, 'RED')# create ball object
    ball2 = disk (100 , 100,  10, 'BLACK')# create ball object

    xInc1 = 2; yInc1 =  2                # set increments for movement
    xInc2 = 4; yInc2 = -1                # set increments for movement
    for n in range(100):                 # start the loop
        canvas.move(ball1, xInc1, yInc1) # move ball both horiz. and vert.
        canvas.move(ball2, xInc2, yInc2)
        canvas.update()                  # update the screen
        sleep(0.01)                      # pause in seconds
#-------------------------------------------------------------------------------------------Tk Graphics demo--
#--13.ANIMATION--Type II. In this form of animation, we use the after() instruction to repeatedly call a move-
#     ment function. As with the previous animation, objects move over (or under) each other. However, with
#     this form of animation, the four loops given below run SIMULTANEOUSLY!

    def move1():
        canvas.move  (ball1, xInc1, yInc1)
        canvas.after(10, move1)
    ball1 = disk (300 , 300,  10, 'BLUE')
    canvas.after(0, move1)

    def move2():
        canvas.move  (ball2, -xInc1, yInc1)
        canvas.after(10, move2)
    ball2 = disk (400 , 300,  10, 'WHITE')
    canvas.after(0, move2)




    def move3():
        canvas.move  (ball3, xInc1, -yInc1)
        canvas.after(10, move3)
    ball3 = disk (500 , 300,  10, 'GREEN')
    canvas.after(0, move3)

    def move4():
        canvas.move  (ball4, -xInc1, -yInc1)
        canvas.after(10, move4)
    ball4 = disk (600 , 300,  10, 'CYAN')
    canvas.after(0, move4)
    root.mainloop() # <--Don't forget this line at the bottom, or the graphics will fail.
#====================================================<MAIN>===================================================

def main():
    runGraphicsCode() # Note the imports and global constants listed below.
#-------------------------------------------------------------------------------------------Tk Graphics demo--
if __name__ == '__main__':
#====================================<GLOBAL CONSTANTS and GLOBAL IMPORTS>====================================
#----Tk Graphics and imports
     from tkinter   import Tk, Canvas, PhotoImage, YES, BOTH, NW
     from time      import sleep
     canvas, root  = setUpCanvas()
#----Standard stuff
     from random import random, randint; from math import sqrt; from copy import deepcopy;
     from time import clock; START_TIME = clock(); main(); print('\n      +===<RUN TIME>===+');
     print('      |  %5.2f'%(clock()-START_TIME), 'seconds |'); print('      +================+')
###############################################<END OF PROGRAM>###############################################
