"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                           SORTING                                           ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+

                  You should avoid writing your own sorting code.--P.J. Plauger,
                  Programming on Purpose, Essays on Software Design (Prentice Hall,
                  1993), page 117.


   Python has a built-in MERGE sort accessed by both sorted() and the in-place sort(). The items to be sorted
must all be of the same type.

                  +-----------------------------------------------------------------+
                  | 1. M = sorted(L)  # Returns a sorted list;  L is unaffected.    |
                  |                                                                 |
                  | 2. L.sort()       # List L is sorted in-place. It returns None. |
                  |                                                                 |
                  | DANGER: Do not forget the parentheses or the sorting will fail  |
                  | and you will get no warning messages.                           |
                  +-----------------------------------------------------------------+

   The sorts are ascending (small to large). To change to descending order append reverse = True, like this:

           M = sorted(reverse = True)

           L.sort(reverse = True)

   Lists of lists are sorted by first components, then by second components, then by third, etc. The sub-lists
do not need to all have the same number of components. For example:

                 L = [[2,1, 7], [4,3], [2,5], [4,2], [4,7], [2,0]]
                 L.sort()
                 print (L) # output:[[2, 0], [2, 1, 7], [2, 5], [4, 2], [4, 3], [4, 7]]

   But, perhaps you do not wish to sort on the first item (indexed by 0) in each sub-list. Here is how to
choose another index:

                from operator import itemgetter
                L = [[2,1, 7], [4,3], [2,5], [4,2], [4,7], [2,0]]
                L.sort(key = itemgetter(1))
                print (L) # output: [[2, 0], [2, 1, 7], [4, 2], [4, 3], [2, 5], [4, 7]]]

   How do we sort a queue of objects based on a field called, say fValue?

ANSWER:                  Q.sort(key = lambda i:i.fValue)

   You can sort tuple and string types by first casting them as lists, and then casting them back:

      M = tuple(sorted(list(T)))   # Return a sorted tuple;  T is unaffected.
      T = "".join(sorted(list(S))) # Return a sorted string; S is unaffected.
      T = str(sorted(list(S)))     # ['A', 'B',...] is a list of sorted chars.

   It is possible to call a function on each item before it is sorted--e.g.,

     S = ['good', 'bad', 'EVIL', 'Nice','evil']
     S.sort(key = str.lower, reverse = True)
     print (S) # output: ['Nice', 'good', 'EVIL', 'evil', 'bad']
                   ------------------------------------

     def dist(x):
         from math import sqrt
         return (sqrt(x[0]**2 + x[1]**2 + x[2]**2))
     data = [[2,1, 7], [4,3, 0], [1,2, 3], [3,2,1], [3,1,2], [8,0,0]]
     Lst = sorted(data, key = dist)
     print (Lst) # output:[[1,2,3], [3,2,1], [3,1,2], [4,3,0], [2,1,7], [8,0,0]]
                   ------------------------------------

     data = [[2,1, 7], [4,3, 0], [1,2, 3], [3,2,1], [3,1,2], [8,0,0]]
     from math import sqrt
     Lst = sorted(data, key = (lambda x: sqrt(x[0]**2 + x[1]**2 + x[2]**2)))
     print (Lst) # output:[[1,2,3], [3,2,1], [3,1,2], [4,3,0], [2,1,7], [8,0,0]]
                   ------------------------------------

     x = [[0,0,0], [1,2,3], [3,2,2], [4,2,2], [1,2,0], ]
     x.sort(key = (lambda Lst: sum(Lst)), reverse = True)
     print(x) # output: [[4, 2, 2], [3, 2, 2], [1, 2, 3], [1, 2, 0], [0, 0, 0]]


