"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                    STRINGS and CHARACTERS                                   ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   Why work with strings at all? Why not just work with lists of characters and use all of the built-in list
functions? Answer: Strings have even better built-in functions. But unlike lists, strings are immutable. The
only way to change a string is to create a new string with the same name (memory leak). There are two ways to
do this, using slicing and replace(). So, in effect, immutable strings actually are changeable.

               1. Replace a character with a character:
                  s = 'hello'
                  s = s[:3] + 'X' + s[4:]
                  print ('s =', s)   # output: s = helXo

                2. Replace a character with several characters
                  stng = 'ABCDABCF'
                  s = stng.replace('C', 'xxx')
                  print ('s =', s)   # output: s = ABxxxDABxxxF

                3. Use this replacement function:
                   def change(stng, position, char):
                       return stng[:position] + char + stng[position + 1:]

                  print ('s =', change ('hello', 2, 'X'))  # s = heXlo

   For use of %s (string), %d (integer), %f (float), etc. see the FORMATTING OUTPUT section.

   For strings, double quotes function like single quotes. One advantage of allowing both is that you can now
include one of the symbols (" or ') between the others without resorting to the backslash.

                        print ('She said "I Love you".') # output: She said "I Love you".

   I use " for literature, and ' for programming. The official name of the string type is "str", NOT "string".
Consequently, if you have a variable in your code called "str", other lines of code may not function properly.
String types do not terminate with any special character; their size is kept in memory. The main use for the
string module is the string constants. This section is so long that it has its own index.

There is no char type in Python. In effect, a character is a string of length 1.

            print (ord('A'))     # output 65
            print (chr(65))      # output 'A'
            Note: chr(9679) will print as a solid block   in graphics.
                  chr(9711) will print as a hollow circle in graphics.
                  chr(9899) will print as a solid  circle in graphics.

   Unlike some other languages, strings can be compared by using the relationship operators: ==, !=, <, >,
<=, and >=.



















                           INDEX FOR THIS SECTION
    INDEXING AND SLICING STRINGS (See the section on SLICING).......line 107
    STRING CONSTANTS................................................line 125
    NUMERIC METHODS.................................................line 139
    JOINING and SPLITTING...........................................line 164
    VARIOUS STRING METHODS..........................................line 178
    --dir (listing of methods)......................................line 180
    --copy..........................................................line 190
    --finding the index of a substring..............................line 200
    --appending characters to the end of a string...................line 218
    --extracting a substring........................................line 227
    --deleting a char...............................................line 231
    --change a character inside of a string.(replace)...............line 235
    --cast a string into a list of individual characters............line 240
    --replace (great for a hangman game)............................line 247
    --alphabetizing the characters in a string......................line 285
    --counting substrings (frequency distributions).................line 290
    --concatenate (using +=)....... ................................line 294
    --concatenate integers into a big string........................line 313
    --casting to a different data types.............................line 324
    --converting from int to string to list and back (useful).......line 230
    --ordinal values of characters..................................line 350
    --comparing strings (using ==, !=, <, >, <=, and >=)............line 360
    --iteration and membership......................................line 367
    --upper and lower case conversions..............................line 372
    --boolean tests for strings.....................................line 377
    --stripping.....................................................line 389
    --capitalize....................................................line 396
    --length of string..............................................line 403
    --reverse and rotate a string...................................line 409
    --eval..........................................................line 450
    --regular expressions (findall).................................line 448
    --Listing individual characters in a string.....................line 456
    --permutations of characters (a list of tuples).................line 463
"""
#==============================SYSTEM COMMNANDS ===============================
from string    import * # punctuation symbols, etc.
def main():
#                         INDEXING AND SLICING STRINGS
   stng = 'ABCDEFG'
   print (stng[ 2])    # output: C
   print (stng[-3])    # output: E
   print (stng[1:2])   # output: B
   print (stng[1:1])   # output: (no output)
   print (stng[2:4])   # output: CD
   print (stng[0:4])   # output: ABCD
   print (stng[:4])    # output: ABCD
   print (stng[3:7])   # output: DEFG
   print (stng[3:])    # output: DEFG
   print (stng[:])     # output: ABCDEFG
   print (stng[:])     # output: ABCDEFG
   print (stng[-4:-2]) # output: DE
   print (stng[-4:])   # output: DEFG
   print (stng[-4:-1]) # output: DEF

#######################################=#######################################
#                           STRING CONSTANTS (useful)
#--Constant strings (HANDY!).
#  from string import *
   print(punctuation)    # output: !"#$%&'()*+,-./:;<=>?@[\]^_`{|}~ (32 chars)
   print(digits)         # output: 0123456789 (10 chars)
   print(hexdigits)      # output: 0123456789abcdefABCDEF
   print(octdigits)      # output: 01234567
   s = whitespace        # newlines, tabs, spaces, etc. (6 unprintable chars)
   print(ascii_letters)  # abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ
   print(ascii_lowercase)# output: abcdefghijklmnopqrstuvwxyz
   print(ascii_uppercase)# output: ABCDEFGHIJKLMNOPQRSTUVWXYZ
   print(printable[0:95])# digits + ascii_letters + punctuation

#######################################=#######################################
#                                NUMERIC METHODS
#--Extract 3rd digit from RIGHT end of integer.
   n = 123456
   print('digit =', n//100%10)       # output: digit = 4. Method 1: using \\ and %.
   print('digit =', int(str(n)[-3])) # output: digit = 4. Method 2: casting to a string type.

#--Extract 3rd digit from LEFT end of integer.
   n = 123456
   print('digit =', int(str(n)[2])) # output: digit = 3

#--Determine how many digits are in an INTEGER.
   from math import floor, log10
   n = -1234567
   print ('length =', len(str(abs(n))))       # Method 1. output: length = 7
   print ('length =', floor(log10(abs(n))+1)) # Method 2. output: length = 7

# modify nth digit from left side of an integer
   num = 123456
   n = 3
   digit = 7
   stng = str(num)
   ans = int(stng[0:n] + str(digit) + stng[n+1:])
   print (ans) # output: 123756

#######################################=#######################################
#                            JOINING and SPLITTING
#--Join (inverse of split).
   Lst = ['1', '2', '3', '4']      # In Python we may join only string types.
   print ('Ans = ', ''.join(Lst))  # Ans = 1234, a string type
   sep = '+'
   print ('Ans = ', sep.join(Lst)) # Ans = 1+2+3+4, a string type

#--Split (inverse of join).
   print('inverse = ', '1+2+3+4'.split('+')) # inverse =  ['1', '2', '3', '4']
   stng = 'to Have and Have not'
   print('stng =', stng.split()) # stng = ['to', 'Have', 'and', 'Have', 'not']
   print('stng =', stng.split('Have')) # stng = ['to ', ' and ', ' not']

#######################################=#######################################
#                           VARIOUS STRING METHODS
#--Print methods list. # prints all string functions (See below.)
    print (dir(str)) #<--Lists all built-in functions
['__add__', '__class__', '__contains__', '__delattr__', '__dir__', '__doc__', '__eq__', '__format__',
'__ge__', '__getattribute__', '__getitem__', '__getnewargs__', '__gt__', '__hash__', '__init__', '__iter__',
'__le__', '__len__', '__lt__', '__mod__', '__mul__', '__ne__', '__new__', '__reduce__', '__reduce_ex__',
'__repr__', '__rmod__', '__rmul__', '__setattr__', '__sizeof__', '__str__', '__subclasshook__', 'capitalize',
'casefold', 'center', 'count', 'encode', 'endswith', 'expandtabs', 'find', 'format', 'format_map', 'index',
'isalnum', 'isalpha', 'isdecimal', 'isdigit', 'isidentifier', 'islower', 'isnumeric', 'isprintable',
'isspace', 'istitle', 'isupper', 'join', 'ljust', 'lower', 'lstrip', 'maketrans', 'partition', 'replace',
'rfind', 'rindex', 'rjust', 'rpartition', 'rsplit', 'rstrip', 'split', 'splitlines', 'startswith', 'strip',
'swapcase', 'title', 'translate', 'upper', 'zfill']
###--Copy.
   stng1 = 'aaaa'
   stng2 = 'bbbb'
   stng1 = stng2           # These are still two different places in memory.
   stng2 = '666'
   print (stng1, stng2)    # output: bbbb 666
   stng1 = stng2[:]        # These are still two different places in memory.
   stng2 = '777'
   print (stng1, stng2)    # output: 666 777

#--Finding the index of a substring.
   stng = 'twotwotwo'




#--Method 1 (use find())
   print('pos = ', stng.find('two'))       # pos =  0
   print('pos = ', stng.find('zzz'))       # pos = -1. This means "not found".
   print('pos = ', stng.find('two', 1))    # pos =  3, with starting position = 1
   print('pos = ', stng.find('two', 1, 4)) # pos = -1, with start and end positions given

#--Method 2 (use index())
   print('pos = ', stng.index('two'))       # pos =  0
#  print('pos = ', stng.index('zzz'))       # run-time ERROR
   print('pos = ', stng.index('two', 1))    # pos =  3, with starting position = 1
#  print('pos = ', stng.index('two', 1, 4)) # run-time ERROR

#  Strings are immutable. To append a new character to the end of a string  (s += 'x'), the entire string must
#  be reconstructed. For repeated appending, work with a list (L) of characters, and then join the characters
#  together into a single string: s = ''.join(L).
   stng = 'ABCDEF'
   Lst = list(stng) + ['G']
   print ('list = ', Lst)    # list =  ['A', 'B', 'C', 'D', 'E', 'F', 'G']
   stng = ''.join(Lst)
   print ('string = ', stng) # string = ABCDEFG

#--Extract a substring.
   s   = 'ABCDEF'
   print(s[2:4]) # output: CD

#--Delete a char.
   s   = 'ABCDEF'
   print(s[:3]+s[4:]) # output: ABCEF

#--Change a character inside of a string. But why not use the built-in replace?
   s = 'hello'
   s = s[:3] + 'X' + s[4:] # output: s =  helXo
   print ('s = ', s)

#--cast a string into a list of individual characters
   s = 'abcd'
   t = list(s)
   print (t)    # output: ['a', 'b', 'c', 'd']
   r = ''.join(t)
   print (r)    # output: abcd

#--Replace (great for a hangman game). Remember a translate(maketrans()) is used for multiple replacements.
   s = 'hello'
   s = s.replace('l', 'xx')
   print (s)                # output: hexxxxo

#  If the replacements are all blanks (' '), then split() can be used to delete the blanks.
   stng = 'to Have    and Have not'
   s = stng.replace('Hav', 'xxx')
   print (s)            # output: to xxxe    and xxxe not
   s = s.split()
   print ('s =', s)     # output: ['to', 'xxxe', 'and', 'xxxe', 'not']
   Ans = ''
   for elt in s:
      Ans += (elt + ' ') # You cannot use append() for strings.
   print ('Ans =', Ans) # output: Ans = to xxxe and xxxe not











   #-replace all 32 punctuation marks and 6 whitespaces with blank spaces. Python 3.x requires "str.maketrans"
   # versus "maketrans()" in Python 2.x.
   s = "one.two;:'  three!?'" # sample string.
   import string
   s = s.translate( str.maketrans( string.punctuation, ' '*32))
   s = s.translate( str.maketrans( string.whitespace , ' '*6 ))
   # or both lines can be combined, as given below
   s = s.translate( str.maketrans(string.punctuation+string.whitespace, ' '*38))
   print ('s =', s) # output: s = one two     three
   s = s.split()
   print ('s =', s) # output: s = ['one', 'two', 'three']

#--Alphabetizing the characters in a string. Note: sorted() returns a sorted sequence, but ".sort()" sorts in
#  place and does NOT return a sequence.
   sl = sorted(list('BCABBED'))
   print ('Ans =', ''.join(sl)) # output: Ans = ABBBBCDE

#--Count substrings. (great for word frequencies)
   stng = 'xabxxxabxxab'
   print('count =',  stng.count('ab')) # output: count = 3

#--Concatenate. [You can NOT use append() with an immutable string.]
   stng1 = 'ABC'; stng2 = 'DEF'; stng3 = 'GHI'

   # Method 1 (slow)
   stng = 'X' + stng1 + stng2
   print ('stng =', stng) # output: stng = XABCDEF

   # method 2 (slow)
   stng = 'A' "BC" 'D'    # implicit concatenation of literals
   print ('stng =', stng) # output: stng = ABCD

   # method 3 (fast)
   print ('concats = ', ''.join(  [stng1, stng2, stng3])) # concats =  ABCDEFGHI
   print ("concats = ", ', '.join([stng1, stng2, stng3])) # output: concats =  ABC, DEF, GHI

   # iteration concatenation
   n = 50
   print ('-' * n) # output: --------------------------------------------------

#--Concatenate integers into a big string.
   # Method 1
   Lst = []
   for num in range(15):
      Lst.append(str(num))
   stng = ''.join(Lst)
   print (stng)                                    # output: 01234567891011121314

   # Method 2
   print(''.join([str(num) for num in range(15)])) # output: 01234567891011121314

#--Cast to a different data type.
   print (str  ( -12.34 )) # float (or int) cast to string
   print (float('-12.34')) # string cast to float
   print (int  ('-12'   )) # string cast to int
   print (int  (-12.34  )) # float  cast to int

#--Converting from int to string to list and back (useful):
   stng = str(9123)[1:]     # convert integer to string (omit leading digit)
   print  ('string: ', stng)# '123'

   num  = int('9' + stng)   # convert string to integer (append leading digit)
   print  ('integer:', num) # 9123

   Lst  = list('123')       # convert string to list
   print  ('list:', Lst)    # ['1', '2', '3']


   stng = ''.join(Lst)      # convert list to string
   print  ('string: ', stng)# '123'

   num  = int('9' + ''.join(Lst)) # convert list to integer and append digit
   print  ('integer:', num) # ['1', '2', '3']

   Lst  = list(str(123))    # convert integer to list
   print(Lst)

#--Ordinal values.
   print (ord('A')) # inverse of chr. output: 65
   print (chr( 65)) # inverse of ord. output: A

   #-Build a string from a list of ord values.
   print(''.join(map (chr, [65, 66, 67, 68]))) # output: ABCD

   #-Turn a string into a list of ord values.
   print(list(map (ord, 'ABCD'))) # output: [65, 66, 67, 68]

#--compare strings.  (Strings can be compared with the standard operators: ==, !=, <, >, <=, and >=.
   str1 = 'cat'; str2 = 'dog'
   if str1 < str2:
      print(str1, 'is less than', str2) # output: 'cat' is less than 'dog'
   if str1 != str2:
      print(str1, '!=', str2)           # output: 'cat' != 'dog'

#--Iteration and membership.
   for elt in list ('ABCD'):
      print (elt, end = ', ')   # output: A, B, C, D,
   print ('\n', 'BC' in 'ABCD') # output: True

#--Upper and lower case conversions.
   stng = 'HAPPY and SAD.'
   print ('lower = ', stng.lower()) # output: lower =  happy and sad.
   print ('upper = ', stng.upper()) # output: upper =  HAPPY AND SAD.

#--Boolean tests for strings.
   print ('1.', '1234'.isdigit()) # output: True
   print ('2.', '1234'.isalpha()) # output: False
   print ('3.', 'abCD'.isalpha()) # output: True
   print ('4.', 'ab D'.isalpha()) # output: False (The space makes it false.)
   print ('5.', '1234'.isalnum()) # output: True alnum = alpha-numeric
   print ('6.', '1aBC'.isalnum()) # output: True
   print ('7.', '-1.0'.isalnum()) # output: False
   print ('8.', '12 C'.isalnum()) # output: False (The space makes it false.)
   print ('9.', 'abcd'.islower()) # output: True
   print ('0.', 'abcd'.isupper()) # output: False

#--Stripping. (Yeah, I know.) Useful for reading file lines. Recall whitespace is six characters, including
#  '\n', ' ', and tab.
   stng = '  the end.,   '
   print (stng.rstrip())      # strips whitespace-types from RIGHT side.
   print (stng.lstrip())      # strips whitespace-types from LEFT  side.
   print ('*', stng.strip('., '), '*' )  # output: * the end *

#--Capitalize.
   stng = 'heLP Me pLeaSe.'
   print (stng.capitalize())        # output: Help me please
   stng = capwords(stng)
   print ('stng =', stng)           # output: Help Me Please.
   print ('stng =', stng.swapcase())# output: hELP mE pLEASE.

#--Length of string.
   print (len('abcdef'))        # length of string




#--Reverse and rotate a string.
   s = 'ABCDEF'
   s = ''.join(reversed(s))  # method 1.
   print (s)

   s = 'ABCDEF'
   s = s[::-1]               # method 2. (better)
   print (s)

   s = 'ABCDEF'
   t = ''
   for n in range(len(s)-1, -1, -1):
         t += s[n]          # method 3. (terrible)
   s = t

   #-Reverse by characters.
   s = 'ABCDEF'
   print (''.join(reversed(s))) # method 1.
   print (s[::-1])              # method 2. (better)

   #-Rotate front of string to the back.
   s = 'ABCDEFG'
   i = 2
   print(s[i:] + s[:i])# output: CDEFGAB

   #-Reverse by words.
   s1 = 'SEND MORE MONEY'
   s2 = s1.split()
   s2.reverse()
   s2 = ' '.join(s2)
   print('s2 =', s2) # output: s2 = MONEY MORE SEND

#--The amazing eval() function. Note: The string evaluated and executed could be a direction to delete files.
   print ( 'Ans1 =', eval('1 + 1 == 2') ) # output: Ans1 = True
   print ( 'Ans2 =', eval('1 + 1') )      # output: Ans2 = 2
   print ( 'Ans3 =', eval("'a' + 'b'") )  # output: Ans3 = ab
   a = 1; b = 2; c = 3
   print ( 'Ans4 =', eval('(a + b) == c') ) # output: Ans3 = True

#--Regular expressions (slightly).
   from re import findall
   print (findall('[0-9]+', '16 2-by-4s in rows of 8'))
   # output: ['16', '2', '4', '8']

   print(findall('[A-Z]+', 'HELLO - NICE == DAY'))
   # output: ['HELLO', 'NICE', 'DAY]

#--Listing individual characters in a string.
   stng = 'send + more == money'
   from string import ascii_lowercase
   chars = list(set(stng).intersection(ascii_lowercase))
   print ('chars = ', chars) # output: chars =  ['d', 'e', 'm', 'n', 'o', 'r', 's', 'y']
   chars = ''.join(chars)    # output: chars =  demnorsy

#--Permutations of characters (a list of tuples) with a built-in method.
   from itertools import permutations
   print(list(permutations('ABC', 3))) # [('A', 'B', 'C'), ('A', 'C', 'B'), etc.

#------------------------------------------------------------------------------
if __name__ == '__main__': main()
#=============================== END OF PROGRAM ===============================
