""" +===============================+========-========*========-========+===============================+
    ||                                     TRANSLATE and MAKETRANS                                     ||
    +===============================+========-========*========-========+===============================+


   The maketrans() function makes a translation table (a DICTIONARY), which is used by the translate()
function to map one character to another. The translate() function is a multiple replace statement for a
string. Alas, The translate() function only works with the ordinal values of the characters
(0 <= ord(ch) < OX10000). The ord() function returns the ASCII value of a character. However, the maketrans()
function will make a table for us, allowing us to work with characters, not their ordinal values.

   #--Example 1. You build the dictionary-type table with ordinals
   table = {ord('A'): ord('X'), ord('E'): ord('Y')}
   print("table =", table)                   # output: table = {65: 88, 69: 89}
   print('AEIOU'.translate(table))           # output: XYIOU

   #--Example 2. build the table with a for loop and zip
   keys   = "aeiou"
   values = "12345"
   print('aeeiiiux'.translate(dict((ord(x),y) for (x,y) in zip(keys, values))))
                                             # output from above: 1223335x

   #--Example 3. maketrans() builds the table for you (the easy way). Python 3.x requires "str.maketrans"
   #             versus "maketrans()" in Python 2.x.
   keys   = "aeiou"
   values = "12345"
   table  = str.maketrans(keys, values)
   print("table =", table)       # {97: 49, 111: 52, 117: 53, 101: 50, 105: 51}
   print('aeeiiiux'.translate(table))# output: 1223335x
   #-working with bytes to do the same operation
   keys  = b"aeiou"
   values= b"12345"
   table = bytes.maketrans(keys, values)
   strg = b"aeeiiiux"
   print(strg.translate(table))       # output: b'1223335x

   #--Example 4. This is an application often associated with files. replace all 32 punctuation marks and 6
   #             whitespaces with blank spaces.

   s = "one.two;:'  three!?'" # sample string.
   import string # necessary for the string constants
   s = s.translate( str.maketrans( string.punctuation, " "*32))
   s = s.translate( str.maketrans( string.whitespace , " "*6 ))
   # or both lines can be combined, as given below
   s = s.translate( str.maketrans(string.punctuation+string.whitespace, " "*38))
   print("s =", s)         # output: s = one two     three
   print("s =", s.split()) # output: s = ['one', 'two', 'three']