"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                     THE PYTHON TURTLE                                       ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   The Python language includes a graphics sub-language called Turtle graphics. Turtle graphics are useful
for a narrow set of problems: drawing mathematical designs.

   The Python Turtle is a drawing pen with a direction. Thus, we can talk about the pen's left, etc. The
turtle pen's initial position defaults to the center of the Turtle window (x=0, y=0), and faces east (heading
= 0 degrees). Angular directions are initially set to degrees. Circles are drawn counterclockwise, unless set
differently. There are 80+ turtle commands. We list the most useful ones below. The others can be found on
the Internet, along with the names of hundreds of Python colors.

   WARNING: Do NOT name your program turtle, or random, or any reserved word. If you do, you will have to
rename your program(s) and remove any copies of the programs that are named with the reserved names.

 1. color("black") # The default pen color is black. Consequently, there is some advantage to setting the
                     background color to black:  Initially, you can move the pen without bothering to lift
                     it up. There are oodles of drawing colors. The most common colors are red, blue1234,
                     cyan1234, grey1234...100, green, magenta1234, orange1234, yellow1234, black, brown1234,
                     etc. The colors blue and blue4 are different shades of blue. Yes, there are 100 shades of
                     grey (not spelled "gray").
 2. bgcolor("blue")# sets the background color
 3. width()        # sets thickness of line. 1 = normal (one pixel), 2 is noticeably thicker
 4. up()           # lifts the drawing pen up, probably to move the pen somewhere else, without leaving a line
 5. down()         # drops the pen down for drawing.
 6. speed(1)       # speed(1) is the slowest and is also the default speed for drawing. speed(5) is faster.
                     speed(10) is very fast, speed(0) is the fastest of all (strange numbering there).
                     Circles are always slower to draw than lines.
 7. hideturtle()   # This hides the turtle shape, and speeds up the drawing.
 8. showturtle()   # This makes the turtle visible again.
 9. forward(10)    # draw straight line 10 pixels forward
10. backward(10)   # draw straight line 10 pixels backward
11. drawLine(a,b,c,d) # This necessary function does NOT exist. Why? So we
                     create it.

                   +----------------------------------------------+
                   | def drawLine(a,b,c,d): # from (a,b) to (c,d) |
                   |            up()                              |
                   |            goto(a,b)                         |
                   |            down()                            |
                   |            goto(c,d)                         |
                   +----------------------------------------------+

12. goto(x,y)      # goto position (x,y). (0,0) is the center of the window, unless you reset it. The goto()
                     instruction will draw on the Turtle's window unless the pen is up, or the pen is set to
                     the background color.
13. right(120)     # turn right 120 degrees (degrees are the default)
14. left (120)     # turn left  120 degrees (degrees are the default)

15. setheading(90) # sets the pen's heading 0 = east, 90 = north, 180 = west, and 270 = south if in the degree
                   # mode.
16. position()     # returns the turtle's current position
17. heading()      # returns the Turtle's current heading
18. towards(x, y)  # points the turtle towards the given point (x, y).
19. circle (50)    # draws a circle with center 50 pixels to the LEFT of the turtle pen. I prefer to write my
                   # own circle function:

  +---------------------------------------------------------------------------+
  | def circlexy(x, y, r):  # This circle function uses the built-in circle().|
  |   r = abs(r)            # The drawing begins in the 3:00 position.        |
  |   up()                  # Lift the pen off the canvas.                    |
  |   goto (x+r, y)         # We wish the circle's center to be at (x,y).     |
  |   setheading(90)        # 90 = face north.                                |
  |   down()                # Put the pen down on canvas.                     |
  |   circle(r)             # Call the built-in circle() command.             |
  +---------------------------------------------------------------------------+

20. dot("blue")    # draws a blue dot of default size (dia. = 4 pixels). But dot(10, "blue") produces a dot of
                   # dia. = size 10 pixels.
21. begin_fill()   # Calling begin_fill() and later end_fill() will color in a shape constructed between the
                   # calls.
22. fillcolor("red")# The color to fill a shape. Used with begin_fill() and
                      end_fill()
23. end_fill()     # See begin_fill().
24. radians()      # sets the turtle's angular measurements to radians
25. degrees()      # sets the turtle's angular measurements to degrees (default)
26. stamp()        # stamps a turtle shape on the canvas.
27. write("X =")   # Writes text to screen. Here is the complete function:

    write("GOODBYE", move=False, align="left", font=("Arial", 20, "normal"))

28. onclick(doit(), btn=1, None)
                   # A left button click (= btn 1) will execute a two-parameter function--e.g.,  doit(a,b).
                   # Two parameters are required. Use onclick (None) to turn off the mouse. The functions
                   # nrelease() and ondrag() are also available.
29. reset()        # clears the window and resets the pen to (0,0)
30. sleep(5)       # pauses execution by 5 seconds
31. clear()        # clears the turtle window
32. exit()         # If you use exit(), then proceed it with sleep (5).

33. setworldcoordinates(0,0,100,100)
                   # (0,0) is lower left, and (100,100) is upper right. This sets the turtle window to be the
                   # first quadrant

34. setup()        # The default turtle window setup is

    setup(width = 0.75, height = 0.5, start x = None, starty = None)

                   # Curiously, you may need to use

                                  setup(1000,1000, None, None)

                     The width of the turtle window is 0.75 of the screen. The height is 0.5 of the screen.
                     None means place the pen on the center of the indicated axis.

35. title("stng)   # Here is how to center a title in the window:

                     title((" "*80 + "===-TURTLE GRAPHICS by M. Stueben-==="))

36. bye()          # If you are having trouble with your windows responding, then try beginning your program
                   # with bye(), which will close the previous window.
                     ------------------------------------

   Also available in turtle are the usual Python functions--e.g., ceil(), cos(), acos(), sin(), asin(), tan(),
atan(), exp(), fabs(), floor(), hypot(x,y) = sqrt(x**2 + y**2), log(), log10(), pow(), and sqrt(). BTW, the
distance function hypot(x,y) = sqrt(x**2 + y**2) is useful in graphical drawings.
"""

















#                              A WORKING PROGRAM
#                        w/ circlexy(), and drawLine())
#================================== IMPORTS ===================================

from math   import sin, cos
from turtle import *
from time   import sleep
PI  = 3.1415926535897932 # This is the maximum accuracy for a float.
#================================= FUNCTIONS ==================================

def circlexy (x, y, r):# This circle function uses the built-in circle().
    r = abs(r)         # The drawing begins in the 3:00 position.
    up()               # lift pen off canvas
    goto (x+r, y)      # we wish the circle's center to be at (x,y)
    setheading(90)     # in DEGREES 90   = face north. Choose one.
 #  setheading(PI/2)   # in RADIANS PI/2 = face north. Choose one.
    down()             # put pen down on canvas
    circle(r)          # call the built-in circle() command
#------------------------------------------------------------------------------

def drawLine(a, b, c, d): # from (a,b) to (c,d)
    up()
    goto(a, b)
    down()
    goto(c, d)
#------------------------------------------------------------------------------

def main():
    r = 100                 # r = base circle radius
    radians()               # change from default degrees
    MAX = 2*PI
    t = 0                   # t = angle theta, the step
    bgcolor('bLACK')
    color('YELLOW')
    width(3)
    drawLine(-100,-100,  100,  100  ) # Turtle window is not quite centered

    circlexy (0, 0, 200)
    #--------------------------
    color('RED')
    while t < MAX:
        x = int(r*cos(t))
        y = int(r*sin(t))
        circlexy(x, y, y)
        t += 0.1
    sleep(2)                # <-- Import the time module.

#-------------------------------------------------------------------------------------------------------------
if __name__ == '__main__':
     from random import random, randint; from math import sqrt; from copy import deepcopy;
     from time import clock; START_TIME = clock(); main(); print('\n      +===<RUN TIME>===+');
     print('      |  %5.2f'%(clock()-START_TIME), 'seconds |'); print('      +================+')
###############################################<END OF PROGRAM>###############################################























"""
   ASSIGNMENTS.
      1) Copy and run Drawing 1 (nephroid), below. The other drawings are only
         for reference, but might be worth your time to copy and run.

      2) Draw a cardioid (heart-shaped) image. Like the nephroid, we draw a
         bunch of circles whose centers are equally spaced on a big base circle.
         The other circles all pass through the same point on the base circle.
         The cardioiod is the envelope (approximated by the boundary) of the
         circles.

      3) Draw an image of your own creation.

      4) Draw the Koch snowflake (first published in 1904) with a recursive
         Turtle function. This is where Turtle graphics show their value. See
         handout.
         """