"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                         TYPE TESTING                                        ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   In a dynamically typed language the computer knows the type of any data or object it manipulates. However,
sometimes the programmer needs to make a decision based on what type is returned from or passed to a function.
We have two tools that will give us the data type:

                                            type() and isinstance().

Example 1. See the program below.
"""
def main():
    x = 5
    y = 3.1
    if type(x) == int:
        print ('1.',True) # output: 1. True

    if type(x) in (int, float, tuple, list, set, str):
        print ('2.',True) # output: 2. True

    if isinstance(x, (int, float)):
        print ('3.',True) # output: 3. True

    if type(x) != type(y):
        print ('4.',True) # output: 4. True

    print(type(1))               # output: <class 'int'>
    print(type(1.0))             # output: <class 'float'>
    print(type('ABC'))           # output: <class 'str'>
    print(type([1,2]))           # output: <class 'list'>
    print(type((1,2)))           # output: <class 'tuple'>
    print(type({1,2}))           # output: <class 'set'>
    print(type({'A':1, 'B':2,})) # output: <class 'dict'>
    print(type(int))             # output: <class 'type'> <--curious
    print(type(type))            # output: <class 'type'> <--curious

if __name__ == '__main__': main()
#-------------------------------------------------------------------------------------------------------------

Example 2. Suppose you wished to flatten a list into a single list. In other words, turn L into Lst:

        L   = [1,[2,3],4, [5, [6, 7]], [8, 9, [1,2, [3,4] ] ] ],

        Lst = [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3, 4].

   The following recursive function checks the type of elements in the given list (L) as it builds the new
list (Lst).

    def flatten (L):
      Lst = []
      for elt in L:
         if type(elt) in (list, tuple):
               Lst.extend(flatten(elt))
         else: Lst.append(elt)
      return Lst

    L = [1,[2,3],4, [5, [6, 7]], [8, 9, [1,2, [3,4] ] ] ]
    print ('7.', flatten (L)) # output: [1, 2, 3, 4, 5, 6, 7, 8, 9, 1, 2, 3, 4]
#-------------------------------------------------------------------------------------------------------------

Example 3. There are several kinds of vector multiplication (scalar times vector, vector times scalar, matrix
           times column vector, row vector times matrix, dot product, and cross product). Here is one way to
           overload the * operators for vectors:

  def __mul__(self, entity): # Overload "*" in 2 forms: V*s, and V.crossProd(W).
     if isinstance(entity, (Vector)):
       return self.crossProd(entity) # <--Get value from the crossProd() method.
     if isinstance(entity, (int, float)):
       return Vector(*[ j*entity for j in (self.data) ] ) # <-- scalar mult.
     return NotImplemented           # optional

   Note that in s*V the first parameter (s) is NOT a Vector type. Since the * operator is overloaded for
Vectors, only V*s evokes __mul__(self, entity). s*V evokes __rmul__(self, entity).

