"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                     PYTHON DICTIONARIES                                     ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


                                           WHAT IS A DICTIONARY?

   Here are some examples of dictionaries, each with four items.

    Dict = {'A':1, 'B':2, 'C':3, 'D':4}                    # letters associated with positive digits.
    Dict = { 1:'A', 2:'B', 3:'C', 4:'D'}                   # Positive digits associated with letters.
    Dict = {'A':[1,2,3,4], 'B':(5,6), 'D':'George', 'G':4} # A mixture of types.

    Suppose:
         city 'A' connects to city 'B' and city 'C';
         city 'B' connects to city 'A';
         city 'C' connects to city 'A' and city 'D'; and
         city 'D' connects to city 'C'.

    Then we can represent this information as an associative (adjacency) array, shown below.

    Dict = {'A':('B', 'C'), 'B': ('A'), 'C': ('A', 'D'), 'D': ('C')} # Cities with connecting roads.

   A Python dictionary has ITEMS (key-value pairs) separated by commas. Each item is a key
(immutable) followed by a colon, followed by some value. A linguistic dictionary is a list of words (keys)
with definitions (values). A Python dictionary is an UNORDERED collection of keys (usually the immutable
numbers, strings, and tuples) with corresponding values (any data type). A Python dictionary is also a hash
table, and can act like an associative (adjacency) array or associative matrix.

   You use a key to find a value, NEVER the reverse. Consequently, keys must be unique. Values do not need to
be unique. Dictionaries are MUTABLE. You can add items, modify items, and modify values. You can NOT directly
change the immutable key for some value, except by deleting the item (key and value), and then replacing it.
Since keys must be immutable, tuples, but not lists of immutable elements can serve as keys.

   Suppose we needed a huge matrix with mostly zero items. To save on memory, we decide to represent the
matrix by the position-values of the few non-zero elements. A dictionary is ideal for this. In column 10 in
row 5 there is a 7, and in column 10, row 8 there is a 9. On column 20 in row 4 there is a 3, and on column
20, in row 9 there is a 1. All other entries are zeroes. We store and retrieve like this:

    Dict = {(10, 5): 7,
            (10, 8): 9,
            (20, 4): 3,
            (20, 9): 1,}

    Notice the ease of dumping out the dictionary:

    for key in Dict:
        print ('f',key, ' = ', Dict[key], sep = '', end = ', ')

    # output: f(10, 8) = 9, f(10, 5) = 7, f(20, 4) = 3, f(20, 9) = 1,

Dictionaries make great data caches.
#---------------------------------------------------------------------------------------------------
                                        DICTIONARY SYNTAX

   You APPEND new items and MODIFY old items the same way: Dict['C'] = 3.

   You DELETE an item in one of several ways:

   1. del (Dict['Z']), if the item can't be found, an error will be generated.

   2. x = Dict.pop('A'), pops item, and returns the value, not the key, and not the item. If the key is not in
          the dictionary, an error occurs.  However, if you append another parameter, then that parameter will
          be returned instead, thereby  avoiding an error: x = Dict.pop('Z', None).

   3. x = Dict.popitem(), pops and returns a RANDOM item. If the dictionary is empty, an error will be
          generated.

   We can avoid the errors mentioned above by first checking that a key is in a dictionary. We do this with
   the "in" (membership) operator.

                        if 'A' in Dict:
                            x = Dict.pop('A')

   To GET a copy of a value associated with a key, use get():

    Dict = {'A':1, 'B':2, 'C':3, 'D':4}
    print(Dict.get('C')) # output: 3
    print(Dict.get('Z')) # output: None

   Here are some examples.

1. CREATING A DICTIONARY (Example 7 is popular)
   Dict1 = {}     # empty dictionary (set() is an empty set.)
   Dict2 = dict() # empty dictionary
   Dict3 = {'A':1, 'B':2, 'D':3, 'G':4}
   Dict4 = dict.fromkeys(['A', 'B'], 7)       # {'A': 7, 'B': 7}
   Dict5 = {x: x*x for x in range(1, 5)}      # {1: 1, 2: 4, 3: 9, 4: 16} Using a list comprehension.
#  -------------------------------------------------------------------------------------------------

2. CREATING A DICTIONARY WITH ZIP AND TUPLES
    keys   = [x for x in 'ABCD']
    values = list(range(4))
    Dict   = dict(zip(keys, values)) # Zip is useful for dictionary construction.
    print(Dict) # output: {'B': 1, 'C': 2, 'A': 0, 'D': 3} <--Notice the random order.
#  -------------------------------------------------------------------------------------------------

3. PRINTING
   d = {'D':4, 'C':3, 'B':2, 'A':1, }
   print(d         )   # output:              {'A': 1,   'C': 3,   'B': 2,   'D': 4}
   print(d.items() )   # output: dict_items ([('A', 1), ('C', 3), ('B', 2), ('D', 4)]) <--Note the free label.
   print(d.keys()  )   # output: dict_keys  (['A', 'C', 'B', 'D'])
   print(d.values())   # output: dict_values([1, 3, 2, 4])

#--Using a list comprehension, we can go label-free.
   print([i for i in d.items()])  # output: [('A', 1), ('C', 3), ('B', 2), ('D', 4)]
   print([k for k in d.keys()])   # output:  ['A', 'C', 'B', 'D']
   print([v for v in d.values()]) # output:  [1, 3, 2, 4]

#--Sorted printing. Remember dictionaries are unsorted and immutable. So how can they be sorted?
   d = {'A':1, 'B':2, 'G':3, 'D':4}
   dlist = sorted(d.items()) # Sort by key in another memory space.
   for elt in dlist:
       print ("--", elt[0], ":", elt[1])
# output:
#-- A : 1
#-- B : 2
#-- C : 3
#-- D : 4
#  -------------------------------------------------------------------------------------------------

4. LENGTH
   print(len(graph))      # Prints the number of items in the dictionary.
#  -------------------------------------------------------------------------------------------------

5. INSERTING an item, or changing a value for a given key:
   Dict['C'] = 3
#  -------------------------------------------------------------------------------------------------

6. DELETING an item:
    del graph['Z']  # Removes item from dictionary. This causes an error if the item is not in the dictionary.
#  -------------------------------------------------------------------------------------------------




7. MEMBERSHIP
    print ('A' in graph)     # True ("in" replaced ".has_key()" in Python 2.x.)
    print ('A' not in graph) # False
    print (not 'A' in graph) # False
#  -------------------------------------------------------------------------------------------------

8. ITERATING (over keys)
    for x in Dict:
      print (x) # This prints all of the keys.
#  -------------------------------------------------------------------------------------------------

9. COUNTING (Dictionaries are ideal for keeping counts with multiple counters.)
   Dict  = {'A':0, 'B':0, 'D':0, 'G':0}
   stng1 = "ABBD"
   stng2 = "A"
   for elt in stng1:
      if elt not in stng2:
         Dict [elt]+=1
   print (Dict) # output: {'A': 0, 'B': 2, 'D': 1, 'G': 0}
#  ------------------------------------------------------------------------------------------------

10. Keys can be indexed when cast to a list.
    Dict  = {'A':0, 'B':0, 'D':0, 'G':0}
    print(sorted(list(Dict.keys()))[2]) # output: D
#  ------------------------------------------------------------------------------------------------                                 # 3

11. COPY DICTIONARY
    x = deepcopy(Dict)
#  -------------------------------------------------------------------------------------------------

12. REMOVE (all elements)
    x.clear()
#  -------------------------------------------------------------------------------------------------

13. CONSTRUCT NEW DICTIONARY (with same keys, but NO VALUES)
    x = Dict.fromkeys(['G','E'])
    print(x): {'E': None, 'G': None}

"""