"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                         TEXT FILES                                          ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


Here is how to write and then read a couple of strings and numbers to and from a text file.

        #---Create data.
            num1 = 12
            num2 = 34.5
            num3 = 678
            str1 = 'King George'
            str2 = 'Alice'
            str3 = '!,.?*&+-/' # <--Don't include the backslash ('\')

        #---Write both numeric and string data to file.
            fileName1 = 'f:\\stuff.txt'
            file1 = open (fileName1, 'w')  # Do not write 'wb' for 'w'
            file1.write (str(num1) + '\n') # Numbers must be stored as strings.
            file1.write (str(num2) + '\n') # '\n' is the separation character.
            file1.write (str(num3) + '\n')
            file1.write (str1 + '\n')
            file1.write (str2 + '\n')
            file1.close()

        #---Append to the end of a text file.
            file1 = open (fileName1, 'a')
            file1.write  (str3 + '\n')
            file1.close()

        #---Copy one file to another.
            import shutil
            shutil.copyfile('f:\\stuff.txt', 'f:\\junk.txt')

        #---Read both numeric and string data from file.
            fileName2 = 'f:\\junk.txt'
            file1 = open (fileName2, 'r')
            num1 = int  (file1.readline().strip()) # strip removes '\n'.
            num2 = float(file1.readline().strip())
            num3 = int  (file1.readline().strip())
            str1 =       file1.readline().strip()
            str2 =       file1.readline().strip()
            str3 =       file1.readline().strip()
            file1.close()

        #---Print data.
            print('num1 =', num1) # output: num1 = 12
            print('num2 =', num2) # output: num2 = 34.5
            print('num3 =', num3) # output: num3 = 678
            print('str1 =', str1) # output: str1 = King George
            print('str2 =', str2) # output: str2 = Alice
            print('str3 =', str3) # output: str3 = !,.?*&+-/

        #---Read the entire file into one string.
            fileName2 = 'f:\\junk.txt'
            file1 = open (fileName2, 'r')
            line = file1.read()
            file1.close()

        #---Print the data line.
            line = line.replace ('\n', ' ')
            print('line =', line)
            # output: line = 12 34.5 678 King George Alice !,.?*&+-/





        #---Read a particular line from a text file.
            fileName2 = 'f:\\junk.txt'
            file1 = open (fileName2, 'r')
            import linecache
            stng = linecache.getline(fileName2, 3) # Read the 3rd line.
            print ('Third item =', stng) # output: Third item = 678
            file1.close()

        #---Read particular characters from a text file.
            file1 = open (fileName2, 'r')
            print('ten  =', file1.read(10)) # Reads the first 10 chars.
            print('five =', file1.read( 5)) # Reads the next   5 chars.
            file1.close()
            '''output:
               ten  = 12
                34.5
                67
                five = 8
               Kin
            '''

        #---Where are we in the file?
            file1 = open (fileName2, 'r')
            file1.seek(5)         # Go to the 5th character in the file
            print("position =", file1.tell())    # position = 5

        #---How large is the file?
            file1.seek(0,2) # The 2 indicates "from the end of the file"
            print("size in bytes =", file1.tell()) # size in bytes = 46

  TEXT FILES (PICKLE and LOAD). Here is how to write and then read a couple of lists.

        #---Create lists.
            Lst1 = [1,2,3,4, 'A']
            Lst2 = [5,6,7,8, 9.9]

        #---Write lists to file.
            fileName1   = 'f:\\stuff.txt'
            file1 = open (fileName1, 'wb') # Use 'w' in Python 2.x.
            import pickle
            pickle.dump(Lst1, file1)
            pickle.dump(Lst2, file1)
            file1.close() # <-- Don't forget the parentheses.

        #---Read lists from file.
            file1 = open (fileName1, 'rb') # Use 'r' in Python 2.x.
            import pickle
            Lst1 = pickle.load( file1 )
            Lst2 = pickle.load( file1 )
            file1.close() # <-- Don't forget the parentheses.
            print ('Lst1 =', Lst1)
            print ('Lst2 =', Lst2)

       #---Output:
           Lst1 = [1, 2, 3, 4, 'A']
           Lst2 = [5, 6, 7, 8, 9.9]

         WARNINGS:
         1) if a file already exists when you create it, the file is erased. Consequently, work only with a
            copy of an important file.
         2) If you omit the parentheses in the file close function, then the file will not be closed and you
            will get no warnings or error messages.
         3) if you use the single '\' instead of the double '\\', your file may not be recognized. Why?
            Because Python may interpret '\' as part of an escape sequence.
         4) Do not write 'wb' for 'w' or vice versa.



   There are two kinds of Python files: Text files (contain strings) and binary files (contain bytes).
Technically text files also contain only bytes, but these bytes are translated to and from strings for us. We
look only at text files in this section.

   Text files (AKA ASCII or character files) are designed to easily save strings and numbers. Of course, the
numbers must first be converted to strings before they are written to a text file. If you wish to read-and-
write lists and data structures, then you need to use pickle with its dump and load commands. All of this will
make more sense if you look at some examples.

   The simplest way to create a text file is to save a Python file as text or data. I prefer to save it as
.txt, not .py. This is probably the preferred method if you cut-and-paste data from the Internet.

   It is possible to both read and write (= append) to a text file at the same time. WARNING: if a file
already exists when you create it, the file is first erased and then written to. Consequently, work only with
a copy of an important file.

   Writing to a file is buffered. The strings to go into the file are held in memory until the buffer fills
(or the program exits, or you flush the buffer), and only then are the strings actually written to the file.

   Since even a text file is filled with bytes, reading characters requires an encoding scheme. The windows
default is likely to be "cp1252". The Linux default is likely to be "utf 8". When opening to read a file you
should probably make the defaults explicit, like this:

        file1 = open ('C:\en-US.dic', mode = 'r', encoding = 'utf 8')

   But the following will work fine:

        file1 = open ('en-US.dic', 'r')

   I fell into an annoying trap trying to name my file "numbers" (Error: invalid argument "c:\numbers.txt").
The problem was the \n was recognized as an end-of-line character. The fix is to use the string
"c:\\numbers.txt", or choose a file name that does not begin with an "n". Recall that \t means tab, another
trap.

   Above, 'r' means we are opening the file to READ from. If we use 'w' we will WRITE to the file. And, of
course, if the file name does not include a path, then the file is assumed to reside in the current working
directory. We use "\n" as the line-termination character in text files.

   In the first example below, we

   1. create a matrix of random digits,
   2. print the matrix as a check,
   3. create a file (file1) to write to,
   4. store the matrix numbers in file1 (numbers are stored as strings),
   5. close the file,
   6. re-open the file, but for reading this time,
   7. read the numbers (actually strings) back into the matrix.
   8. print the matrix as a check.
"""
#==================================== MAIN ====================================
from random import randint
def main():

   MAX = 5 # = size of square matrix.

#--1. Create a matrix of random digits.-----------------------------------------
   from random import randint # a <= randint(a, b) <= b
   M = [[randint(1, 9) for c in range(MAX)] for r in range(MAX)]
   M[4][4] = 0

#--2. Quick-print the matrix.---------------------------------------------------
   [print(row) for row in M]
   print()
#  output: [2, 8, 6, 2, 4]
#          [4, 2, 5, 4, 8]
#          [4, 7, 3, 8, 6]
#          [5, 5, 6, 7, 1]
#          [3, 2, 3, 7, 0]

#--3. Create a file, named file1, to write to.----------------------------------
   fileX = open ("c:\numberx.dic", mode = "w", encoding = "utf 8")
   file1 = open ("c:\cities.txt", "w")

#--4. Store the matrix in file1. (Numbers are stored as strings.)---------------
   for row in M:
       for elt in row:
           file1.write (str(elt) + '\n') # Numbers are stored as strings.

#--5. Close file1.--------------------------------------------------------------
   file1.close()

#--6. Re-open file1, but for reading this time.---------------------------------
   file1 = open ('c:\cities.txt', 'r')

#--7. Read the numbers (actually strings) back into the matrix.-----------------
   M   = []
   for row in range(MAX):
       line = []
       for col in range(MAX):
           line += [int(file1.readline())] # turn the str back into an int.
       M += [line]

#--8. Print the matrix as a check.
   [print(row) for row in M]
   print()
   if type(M[0][0]) == int: print ('integer') # = a check on the type
#  output: [2, 8, 6, 2, 4]
#          [4, 2, 5, 4, 8]
#          [4, 7, 3, 8, 6]
#          [5, 5, 6, 7, 1]
#          [3, 2, 3, 7, 0]
#
#          integer
   exit('==> Program interupt <==')

#--0. DOWNLOAD and open a real dictionary. Read the file until you find "aardvark". Then print the word. I had
#     to remove all whitespace from each dictionary word before my code would recognize "aardvark". In the
#     systems lab, we have a dictionary (words, no definitions) located at "/usr/share/dict/words". If you are
#     using Windows, then you can down-load an entire free dictionary by Googling "download dictionary file"
#     and going to the Ultimate Spell Dictionary List at

#                           www.karamasoft.com/ultimatespell/dictionary.aspx

    #Read the first ten words
    import string
    string.punctuation ='!"#$%&()*+,-./:;<=>?@[\]^_`{|}~' # apostrophe removed
    table = str.maketrans(string.punctuation+string.whitespace, " "*37)
    file1 = open ("E:\en-US.dic", "r")
    stng = file1.readline().lower().translate(table).strip()
    for n in range (1, 10):
         stng = file1.readline().lower().translate(table).strip()
         print (n, ". *", stng, "*", sep = "")
    file1.close()

    # Search for"aardvark"
    file1 = open ("E:\en-US.dic", "r")
    stng = file1.readline().lower().translate(table).strip()
    while stng != "aardvark":
         stng = file1.readline().lower().translate(table).strip()
    print ("Found:", stng,)
    file1.close()
#===============================================================================




#--1. CREATE a text file (= ASCII file or character file, as opposed to a binary file). Use "\n" as your line-
#     termination character. In other words, open a new (empty) file, write to it, and close the file. The
#     other method of creating a file is to cut-and-paste data into your Python editor and then save the file.
#     WARNING: To avoid accidentally destroying your file, do NOT write to the original file. Also note that
#     if a file already exists when you create it, the file is first erased and then written to. Yikes!
#     Finally, writing to a file is buffered. The strings to go into the file are held in memory until the
#     buffer fills (or the program exits, or you flush the buffer), and only then are the strings actually
#     written to the file.

    #-Method 1. Use file.write().
    file2 = open ("c:\stuff.txt", "w") # "w" for "write to the file".
    file2.write  ("This is line 1.\n")
    file2.write  ("This is line 2.\n")
    file2.close()                      # Don't forget to close.

    #-Method 2. Use file.writelines(). Lines are separated by "\n", which the
    # programmer must place in each of his lines.
    Lst =["--This is line 1.\n", "--This is line 2.\n"]
    file2 = open ("c:\stuff.txt", "w") # "w" for "write to the file".
    file2.writelines  (Lst)
    file2.close()

#--2. REPLACE text in a file. (This is useful.)
    file1 = open ("c:\poem.txt", "r+") # "r+" for "read and write to the file".
    file1.write(file1.read().replace("Mary", "MaryMary"))
    file1.close()

#--3. APPEND to the end of a text file.
    file1 = open ("c:\poem.txt", "a")
    file1.write  ("This is an APPENDED line to the text.\n")
    file1.close()

#--4. Copy one file to another.
    import shutil
    shutil.copyfile("G:\poem2.txt", "c:\poem.txt") # from 1st name to 2nd name.

#--5. READ an ENTIRE text file into one string, and then print it. Note: You may prefer a list of strings
#    instead. See below. Also, the "r" for read is also the default.
    file1 = open("c:\poem.txt", "r")   # "r" for "read from the file".
    stng = file1.read()
    print (stng)
    file1.close()

#--6. READ an ENTIRE text file into a list of strings, and then print them.
    file1 = open("c:\poem.txt", "r")   # "r" for "read from the file".
    stngs = file1.read().split('\n')  # This is a list of strings.
    print (stng)
    file1.close()

#--6. READ an ENTIRE text file, line by line.
    file1 = open("c:\poem.txt", "r")
    for line in file1:
       line.rstrip()                   # <--strips off "\n", & other whitespace.
       print (line, end = "")
    file1.close()

#--7. Read an entire text file into a LIST OF STRINGS. (This is useful.)
    file1 = open ("c:\poem.txt", "r")
    lines = file1.readlines()
    print (lines[0], lines[1], sep = "") # This prints the first two lines.
    file1.close()

#--8. READ one LINE at a time from a text file.
    file1 = open ("c:\poem.txt", "r")
    print (file1.readline(), end = "") # This reads the FIRST  line of the file.
    print (file1.readline())           # This reads the SECOND line of the file.
    file1.close()

#--9. READ a particular line from a text file.
    file1 = open("c:\poem.txt", "r")
    import linecache
    stng = linecache.getline("c:\poem.txt", 3) # Read the 3rd line from the file.
    print ("===> ", stng)
    file1.close()

#--10. READ particular characters from a text file.
    file1 = open ("c:\poem.txt", "r")
    print(file1.read(10))            # Reads the first 10 characters from file1.
    print(file1.read(5))             # Reads the next   5 characters from file1.
    file1.close()

#--11. MAKE A LIST of words in a text file, lowercase them, count them, and print them to the screen.
    file1 = open("c:\poem.txt", "r")
    stng = file1.read().lower()
    words = stng.split()
    import string
    table = str.maketrans(string.punctuation+string.whitespace, " "*38)
    for n in range(len(words)):
        words[n] = words[n].translate(table)
        words[n] = words[n].strip()
    print ("There are", len(words), "words in this file.")

    total = []
    for elt in words:
        total.append ((words.count(elt), elt))
    newlist = list(reversed(sorted(set(total))))
    print(newlist)
    print("The three most common words:")
    for elt in newlist[0:3]:
        print (elt[1].ljust(4), elt[0])
    file1.close()
    exit()

#--12. Use seek() and tell() to read a particular character and report the position of the file pointer.
#
#   file1.seek(5, 0) = file1.seek(5) = 5 bytes from front of file.
#   file1.seek(5, 1) = file1.seek(5) = 5 bytes from current file position.
#   file1.seek(5, 2) = file1.seek(5) = 5 bytes from end of file.
#   file1.tell(0 returns the current position of the file pointer.
    file1 = open ("c:\poem.txt", "r")
    file1.seek(5)         # Go to the 5th character in the file
    print(file1.read(10)) # read the 5th to 14th character
    print("position =", file1.tell())    # position = 15
    file1.seek(0)         # return to the beginning of a text file.
    print(file1.read(10)) # output: Mary had a
    file1.close()

#--13. Find the size in bytes) of a file, using seek (to the end) and tell().
    file1 = open ("c:\poem.txt", "r")
    file1.seek(0,2) # The 2 indicates "from the end of the file"
    print("size in bytes =", file1.tell())
    file1.close()

#--14. Transform a text file into a list of strings. This function is often used
#      when processing dictionaries of words.
#
# This function will EXTRACT a list of strings of all equal-length words  from a text file of words of
# different lengths. The function will Lowercase all letters in the words, and strip off all non-alphabetical
# characters especially the end-of-line character ('\n') after each word. It will keep all words of length
# Len. Thus, the 7-character "pull-up" becomes the 6-letter  "pullup", and the 7-character "world's" becomes
# the 6-letter "worlds". The function does NOT allow duplicate words--e.g., "Buster" and "buster" will both be
# extracted to one word, "buster".




def fileToListOfStrings(Len, fileName1):
    from string import ascii_lowercase
    file1      = open (fileName1, 'r')
    dict       = file1.read().lower().split('\n')
    file1.close()
    wordList = []

    for word in dict:
        for ch in word:
            if ch not in ascii_lowercase: # Allow no non-alphabetical chars.
                word = word.replace(ch, '')
        if len(word) != Len:              # All words must be same length.
            continue
        if word not in wordList:          # Allow no duplicate words.
            wordList.append(word)

    return wordList

#--15. Pickle (dump) and unpickle (load) a list. In other words, write a list (or any object) to a BYTE file
#      (this is called pickling) and then read it back from the BYTE file (un-pickling). Unfortunately, we can
#      only load one object to a file. If you have two objects, place them in a list, and pickle the list.
#      Alas, you cannot open and directly read a byte file like you can a text file. Nevertheless, you can
#      open it and print it, or part of it, to the screen.

    Lst1 = [1,2,3,4,]; Lst2 = [5,6,7,8,]
    file1 = open ('c:\stuff.txt', 'wb')
    import pickle
    pickle.dump(Lst1, file1)
    file1.close()

    Lst2 = pickle.load( open( "c:\stuff.txt", "rb" ) ) # Use "r" in Python 2.x.
    print ("Lst2 =", Lst2)
    file1.close()

# If the file already exists, then this is what you want:
    fileName2   = 'c:\stuff.txt'
    file2 = open (fileName2, 'rb')
    import pickle
    data = pickle.load(file2)
    file2.close()
    print (data)

#------------------------------------------------------------------------------
if __name__ == '__main__': main()
#=============================== END OF PROGRAM ===============================
