"""   +=================-=========-=========-=========*=========-=========-=========-=================+
      ||                                  THE PYTHON YIELD STATEMENT                                 ||
      ||                                                                                             ||
      +=================-=========-=========-=========*=========-=========-=========-=================+


   The yield statement is similar to the return statement, with one significant difference. When you re-call a
function that contains yield, the function remembers what it was doing the last time it was called and con-
tinues from that point.

   When you use a yield instead of a return, you have created a generator. You must assign the function name
to another identifer (evidentally the generator is returned), and use __next__() in Python 3.x and next() in
Python 2.x. With "in" you need not reassign. Here are some examples.
"""
def number():
    x = [0,1,2]
    for i in range(len(x)):
        yield(x[i])
#-------------------------=-----------------------=-------------------------=-----------------------
def number(): # This is the same generator as above, only simpler to write.
    yield 0
    yield 1
    yield 2
#-------------------------=-----------------------=-------------------------=-----------------------
def frange(start, stop, step):
    i = start
    while i < stop:
          yield i # <-- not return
          i += step
#-------------------------=-----------------------=-------------------------=-----------------------
def setOfGridPoints():
    from random import random
    for x in range (0, DOMAIN_LIMIT):
       for y in range (0, DOMAIN_LIMIT):
           yield (x + random(), y + random())
#-------------------------=-----------------------=-------------------------=-----------------------
DOMAIN_LIMIT = 10 # Global constant connected to the Nelder-Mead assignment.
def main():
    print(number().__next__()) # output: 0
    print(number().__next__()) # output: 0. You must reassign the function name.
    print()

    x = number()               # this is a reassignment.
    print(x.__next__())        # output: 0
    print(x.__next__())        # output: 1
    print(x.__next__())        # output: 2
   #print(x.__next__())        # output: ERROR, there is nothing left to yield
    print()

    x = number()               # A second reassignment means we begin again.
    print(x.__next__())        # output: 0
    print(x.__next__())        # output: 1
    print(x.__next__())        # output: 2

    from math import pi
    for t in frange (0, 2*pi, 0.1):
        print(round(t, 2), end = ', ')  # output: 0, 0.1, 0.2, 0.3, 0.4, ..., 6.0, 6.1, 6.2,
#-------------------------=-----------------------=-------------------------=-----------------------
if __name__ == '__main__': main()
##########################################<END OF PROGRAM>##########################################

